use std::sync::Arc;

use egui::{
    Align, CentralPanel, Checkbox, Color32, Context, Image, ScrollArea, Slider, TextEdit,
    TextureOptions, Ui, ViewportBuilder, ViewportId,
};
use egui_extras::{Column, Size, StripBuilder, TableBuilder};

use crate::{
    emulator::SimId,
    memory::{MemoryMonitor, MemoryView},
    vram::{VramImage, VramParams, VramProcessor, VramRenderer, VramTextureLoader},
    window::{
        utils::{NumberEdit, UiExt as _},
        AppWindow,
    },
};

use super::utils;

pub struct ObjectWindow {
    sim_id: SimId,
    loader: Arc<VramTextureLoader>,
    objects: MemoryView,
    index: usize,
    generic_palette: bool,
    params: VramParams<ObjectParams>,
    scale: f32,
}

impl ObjectWindow {
    pub fn new(sim_id: SimId, memory: &mut MemoryMonitor, vram: &mut VramProcessor) -> Self {
        let renderer = ObjectRenderer::new(sim_id, memory);
        let ([object], params) = vram.add(renderer);
        let loader = VramTextureLoader::new([("vram://object".into(), object)]);
        Self {
            sim_id,
            loader: Arc::new(loader),
            objects: memory.view(sim_id, 0x0003e000, 0x2000),
            index: 0,
            generic_palette: false,
            params,
            scale: 1.0,
        }
    }

    fn show_form(&mut self, ui: &mut Ui) {
        let row_height = ui.spacing().interact_size.y;
        ui.vertical(|ui| {
            TableBuilder::new(ui)
                .column(Column::auto())
                .column(Column::remainder())
                .body(|mut body| {
                    body.row(row_height, |mut row| {
                        row.col(|ui| {
                            ui.label("Index");
                        });
                        row.col(|ui| {
                            ui.add(NumberEdit::new(&mut self.index).range(0..1024));
                        });
                    });
                });
            ui.section("Properties", |ui| {
                let object = self.objects.borrow().read::<[u16; 4]>(self.index);
                let (mut char_index, mut vflip, mut hflip, palette_index) =
                    utils::parse_cell(object[3]);
                TableBuilder::new(ui)
                    .column(Column::remainder())
                    .column(Column::remainder())
                    .body(|mut body| {
                        body.row(row_height, |mut row| {
                            row.col(|ui| {
                                ui.label("Character");
                            });
                            row.col(|ui| {
                                ui.add_enabled(
                                    false,
                                    NumberEdit::new(&mut char_index).range(0..2048),
                                );
                            });
                        });
                        body.row(row_height, |mut row| {
                            row.col(|ui| {
                                ui.label("Palette");
                            });
                            row.col(|ui| {
                                let mut palette = format!("OBJ {}", palette_index);
                                ui.add_enabled(
                                    false,
                                    TextEdit::singleline(&mut palette).horizontal_align(Align::Max),
                                );
                            });
                        });
                        body.row(row_height, |mut row| {
                            row.col(|ui| {
                                let checkbox = Checkbox::new(&mut hflip, "H-flip");
                                ui.add_enabled(false, checkbox);
                            });
                            row.col(|ui| {
                                let checkbox = Checkbox::new(&mut vflip, "V-flip");
                                ui.add_enabled(false, checkbox);
                            });
                        });
                    });
            });
            ui.section("Display", |ui| {
                ui.horizontal(|ui| {
                    ui.label("Scale");
                    ui.spacing_mut().slider_width = ui.available_width();
                    let slider = Slider::new(&mut self.scale, 1.0..=10.0)
                        .step_by(1.0)
                        .show_value(false);
                    ui.add(slider);
                });
                ui.checkbox(&mut self.generic_palette, "Generic palette");
            });
        });
        self.params.write(ObjectParams {
            index: self.index,
            generic_palette: self.generic_palette,
            ..ObjectParams::default()
        });
    }

    fn show_object(&mut self, ui: &mut Ui) {
        let image = Image::new("vram://object")
            .fit_to_original_size(self.scale)
            .texture_options(TextureOptions::NEAREST);
        ui.add(image);
    }
}

impl AppWindow for ObjectWindow {
    fn viewport_id(&self) -> egui::ViewportId {
        ViewportId::from_hash_of(format!("object-{}", self.sim_id))
    }

    fn sim_id(&self) -> SimId {
        self.sim_id
    }

    fn initial_viewport(&self) -> ViewportBuilder {
        ViewportBuilder::default()
            .with_title(format!("Object Data ({})", self.sim_id))
            .with_inner_size((640.0, 480.0))
    }

    fn on_init(&mut self, ctx: &Context, _render_state: &egui_wgpu::RenderState) {
        ctx.add_texture_loader(self.loader.clone());
    }

    fn show(&mut self, ctx: &Context) {
        CentralPanel::default().show(ctx, |ui| {
            ui.horizontal_top(|ui| {
                StripBuilder::new(ui)
                    .size(Size::relative(0.3))
                    .size(Size::remainder())
                    .horizontal(|mut strip| {
                        strip.cell(|ui| {
                            ScrollArea::vertical().show(ui, |ui| self.show_form(ui));
                        });
                        strip.cell(|ui| {
                            ScrollArea::both().show(ui, |ui| self.show_object(ui));
                        });
                    });
            });
        });
    }
}

#[derive(Clone, PartialEq, Eq)]
struct ObjectParams {
    index: usize,
    generic_palette: bool,
    left_color: Color32,
    right_color: Color32,
}

impl Default for ObjectParams {
    fn default() -> Self {
        Self {
            index: 0,
            generic_palette: false,
            left_color: Color32::from_rgb(0xff, 0x00, 0x00),
            right_color: Color32::from_rgb(0x00, 0xc6, 0xf0),
        }
    }
}

enum Eye {
    Left,
    Right,
}

struct ObjectRenderer {
    chardata: MemoryView,
    objects: MemoryView,
    brightness: MemoryView,
    palettes: MemoryView,
}

impl ObjectRenderer {
    pub fn new(sim_id: SimId, memory: &mut MemoryMonitor) -> Self {
        Self {
            chardata: memory.view(sim_id, 0x00078000, 0x8000),
            objects: memory.view(sim_id, 0x0003e000, 0x2000),
            brightness: memory.view(sim_id, 0x0005f824, 8),
            palettes: memory.view(sim_id, 0x0005f860, 16),
        }
    }

    fn render_object(&self, image: &mut VramImage, params: &ObjectParams, eye: Eye) {
        let chardata = self.chardata.borrow();
        let objects = self.objects.borrow();
        let brightness = self.brightness.borrow();
        let palettes = self.palettes.borrow();

        let object: [u16; 4] = objects.read(params.index);

        let ron = object[1] & 0x4000 != 0;
        let lon = object[1] & 0x8000 != 0;
        if match eye {
            Eye::Left => !lon,
            Eye::Right => !ron,
        } {
            return;
        }

        let brts = brightness.range::<u8>(0, 8);

        let x = ((object[0] & 0x3ff) << 6 >> 6) as i16;
        let parallax = ((object[1] & 0x3ff) << 6 >> 6) as i16;
        let y = ((object[2] & 0x0ff) << 8 >> 8) as i16;

        let (x, color) = match eye {
            Eye::Left => (x - parallax, params.left_color),
            Eye::Right => (x + parallax, params.right_color),
        };

        let (char_index, vflip, hflip, palette_index) = utils::parse_cell(object[3]);
        let char = chardata.range::<u16>(char_index * 8, 8);
        let palette = if params.generic_palette {
            utils::generic_palette(color)
        } else {
            utils::parse_palette(palettes.read(8 + palette_index * 2), brts, color)
        };

        for row in 0..8 {
            let real_y = y + row as i16;
            if !(0..384).contains(&real_y) {
                continue;
            }
            for (col, pixel) in utils::read_char_row(char, hflip, vflip, row).enumerate() {
                let real_x = x + col as i16;
                if !(0..224).contains(&real_x) {
                    continue;
                }
                image.add((real_x as usize, real_y as usize), palette[pixel as usize]);
            }
        }
    }
}

impl VramRenderer<1> for ObjectRenderer {
    type Params = ObjectParams;

    fn sizes(&self) -> [[usize; 2]; 1] {
        [[384, 224]]
    }

    fn render(&mut self, params: &Self::Params, images: &mut [VramImage; 1]) {
        let image = &mut images[0];
        image.clear();
        self.render_object(image, params, Eye::Left);
        self.render_object(image, params, Eye::Right);
    }
}
